from __future__ import annotations

import functools
import inspect
from importlib import util
from typing import Callable, TypeVar, Literal

from absl import logging

R = TypeVar("R")
LogLevel = Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]


def log_before(
    func: Callable[[...], R], logger: Callable[[str], None] = logging.debug
) -> Callable[[...], R]:
    """

    Parameters
    ----------
    func
    logger

    Returns
    -------

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        func_args = inspect.signature(func).bind(*args, **kwargs).arguments
        func_args_str = ", ".join(map("{0[0]} = {0[1]!r}".format, func_args.items()))
        logger(
            f"Entered {func.__module__}.{func.__qualname__} with args ( {func_args_str} )"
        )
        return func(*args, **kwargs)

    return wrapper


def log_after(
    func: Callable[[...], R], logger: Callable[[str], None] = logging.debug
) -> Callable[[...], R]:
    """
    Log's function's return value.

    Parameters
    ----------
    func:
        Function exit from which must be logged.
    logger:
        Logger to use, default absl.logging.debug

    Returns
    -------

    func:
        Function with the same signature.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        retval = func(*args, **kwargs)
        logger(
            f"Exited {func.__module__}.{func.__qualname__}(...) with value: "
            + repr(retval)
        )
        return retval

    return wrapper


def setup_logging(
    *,
    log_format: str = "%(asctime)s:[%(filename)s:%(lineno)s->%(funcName)s()]:%(levelname)s: %(message)s",
    log_level: LogLevel = "DEBUG",
):
    import logging
    import absl.logging

    logging.basicConfig(
        level=logging.getLevelName(log_level),
        format=log_format,
    )

    absl.logging.set_verbosity(absl.logging.converter.ABSL_NAMES[log_level])

    if util.find_spec("tensorflow"):
        import tensorflow as tf

        tf.get_logger().setLevel(log_level)



from __future__ import annotations

import functools
import logging
from typing import Callable, TypeVar
import platform

import jax

R = TypeVar("R")


def requires_gpu(
    func: Callable[[...], R], linux_only: bool = False
) -> Callable[[...], R]:
    """
    Fail if function is executing on host without access to GPU(s).
    Useful for early detecting container runtime misconfigurations.

    Parameters
    ----------
    func:
        Function, which needs hardware acceleration.
    linux_only:
        If set to true, will ignore check on non-linux hosts.


    Returns
    -------

    func:
        Function with the same signature as original one.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        if linux_only and platform.system() != "linux":
            logging.info(
                "Not running on linux, and linux_only==True, ignoring GPU strategy check."
            )
            return func(*args, **kwargs)

        devices = jax.devices()
        logging.info(f"JAX devices -> {devices}")
        if devices[0].device_kind != "gpu":
            raise RuntimeError("No GPU available.")
        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations
from importlib import util


def is_lib_installed(name: str) -> bool:
    return util.find_spec(name) is not None


from absl_extra.src.tasks import run, register_task
from absl_extra.src.notifier import BaseNotifier

if is_lib_installed("slack_sdk"):
    from absl_extra.src.notifier import SlackNotifier
from absl_extra.src.logging_utils import log_before, log_after, setup_logging

if is_lib_installed("pymongo"):
    from absl_extra.src.tasks import MongoConfig
if is_lib_installed("tensorflow"):
    from absl_extra.src import tf_utils

if is_lib_installed("jax"):
    from absl_extra.src import jax_utils



from __future__ import annotations

import json
from functools import wraps
from importlib import util
from typing import Callable, NamedTuple, TypeVar, Mapping, List

from absl import app, flags, logging

from absl_extra.src.notifier import BaseNotifier

T = TypeVar("T", bound=Callable, covariant=True)
FLAGS = flags.FLAGS

if util.find_spec("pymongo"):
    from pymongo import MongoClient
    from pymongo.collection import Collection
else:
    Collection = None
    logging.warning("pymongo not installed.")

if util.find_spec("ml_collections"):
    from ml_collections import config_flags
else:
    logging.warning("ml_collections not installed")


class MongoConfig(NamedTuple):
    uri: str
    db_name: str
    collection: str


class _ExceptionHandlerImpl(app.ExceptionHandler):
    def __init__(self, name: str, notifier: BaseNotifier):
        self.name = name
        self.notifier = notifier

    def handle(self, exception: Exception) -> None:
        self.notifier.notify_job_failed(self.name, exception)


class TaskT(NamedTuple):
    fn: Callable
    name: str | Callable[[str, ...], str]
    init_callbacks: List[Callable[[...], None]] = []


_TASK_STORE = None


def register_task(
    fn: Callable[[str], None],
    name: str | Callable[[], str] = "main",
    init_callback: List[Callable[[...], None]] = None,
) -> None:
    """

    Parameters
    ----------
    fn:
        Function to execute.
    name:
        Name to be used for lifecycle reporting.
    init_callback:
        List of callback, which must be called on initialization.
        By default, will print parsed absl.flags and ml_collection.ConfigDict to stdout.

    Returns
    -------

    """
    global _TASK_STORE
    _TASK_STORE = TaskT(fn=fn, name=name)


def pseudo_main(
    task: TaskT,
    app_name: str,
    notifier: BaseNotifier,
    config_file: str | None = None,
    db: Collection | None = None,
) -> Callable[[str, ...], None]:
    @wraps(task.fn)
    def wrapper(cmd: str):
        if isinstance(task.name, Callable):
            task_name = task.name()
        else:
            task_name = task.name

        logging.info("-" * 50)
        logging.info(
            f"Flags: {json.dumps(flags.FLAGS.flag_values_dict(), sort_keys=True, indent=4)}"
        )

        kwargs = {}

        if util.find_spec("ml_collections") and config_file is not None:
            config = config_flags.DEFINE_config_file("config", default=config_file)
            config = config.value
            logging.info(
                f"Config: {json.dumps(dict(config), sort_keys=True, indent=4)}"
            )
            kwargs["config"] = config
        logging.info("-" * 50)
        if db is not None:
            kwargs["db"] = db
        notifier.notify_job_started(f"{app_name}.{task_name}")
        ret_val = task.fn(cmd, **kwargs)
        notifier.notify_job_finished(f"{app_name}.{task_name}")
        return ret_val

    return wrapper


def run(
    app_name: str | Callable[[], str] = "app",
    notifier: BaseNotifier | None = None,
    config_file: str | None = None,
    mongo_config: MongoConfig | Mapping[str, ...] | None = None,
) -> None:
    """

    Parameters
    ----------
    app_name:
        Name of the task.
    notifier:
        Notifier instance, which monitors execution state.
    config_file:
        Optional, path to ml_collection config file.
    mongo_config:
        Optional, NamedTuple or Dict with MongoDB connection credentials.

    Returns
    -------

    """
    if notifier is None:
        notifier = BaseNotifier()

    if util.find_spec("pymongo") and mongo_config is not None:
        if isinstance(mongo_config, Mapping):
            mongo_config = MongoConfig(**mongo_config)
        db = (
            MongoClient(mongo_config.uri)
            .get_database(mongo_config.db_name)
            .get_collection(mongo_config.collection)
        )
    else:
        db = None

    app.install_exception_handler(_ExceptionHandlerImpl(app_name, notifier))

    task_fn = pseudo_main(
        task=_TASK_STORE,
        notifier=notifier,
        config_file=config_file,
        app_name=app_name,
        db=db,
    )
    app.run(task_fn)



from __future__ import annotations
from absl import logging
from importlib import util


class BaseNotifier:
    def notify_job_started(self, name: str):
        logging.info(f"Job {name} started.")

    def notify_job_finished(self, name: str):
        logging.info(f"Job {name} finished.")

    def notify_job_failed(self, name: str, exception: Exception):
        logging.error(f"Job {name} failed with {exception}")


if util.find_spec("slack_sdk"):
    import slack_sdk

    class SlackNotifier(BaseNotifier):
        def __init__(self, slack_token: str, channel_id: str):
            self.slack_token = slack_token
            self.channel_id = channel_id

        def notify_job_started(self, name: str):
            slack_client = slack_sdk.WebClient(token=self.slack_token)
            slack_client.chat_postMessage(
                channel=self.channel_id,
                blocks=[
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f" :ballot_box_with_check: Job {name} started.",
                        },
                    }
                ],
                text="Job Started!",
            )

        def notify_job_finished(self, name: str):
            slack_client = slack_sdk.WebClient(token=self.slack_token)
            slack_client.chat_postMessage(
                channel=self.channel_id,
                blocks=[
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f":white_check_mark: Job {name} finished execution.",
                        },
                    }
                ],
                text="Job Finished!",
            )

        def notify_job_failed(self, name: str, exception: Exception):
            slack_client = slack_sdk.WebClient(token=self.slack_token)
            slack_client.chat_postMessage(
                channel=self.channel_id,
                blocks=[
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f":x: Job {name} failed, reason:\n ```{exception}```",
                        },
                    }
                ],
                text="Job Failed!",
            )

else:
    logging.warning("slack_sdk not installed.")
